import os
import logging
from threading import RLock
from typing import Callable, List

import tensorflow as tf
from tensorflow.python.client import device_lib

_devices = None
_devices_lock = RLock()


def get_all_devices() -> List[str]:
    """Returns a list of devices in the computer. Example:
    ['CPU:0', 'XLA_GPU:0', 'XLA_CPU:0', 'GPU:0', 'GPU:1', 'GPU:2', 'GPU:3']
    """
    global _devices, _devices_lock

    # Initialize the device list if necessary
    with _devices_lock:
        if _devices is None:
            # Use the TF_FORCE_GPU_ALLOW_GROWTH=true environment variable to
            # force allow tensorflow to take up GPU memory dynamically, instead
            # of allocating 100% of memory at first model-load.
            #
            # TODO: Use tf.config.list_physical_devices in TF 2.1

            with tf.Session():
                all_devices = device_lib.list_local_devices()

            # Get the device names and remove duplicates, just in case...
            tf_discovered_devices = {
                d.name.replace("/device:", "")
                for d in all_devices
            }

            # Discover devices using OpenVINO
            try:
                from openvino.inference_engine import IECore
                ie = IECore()
                openvino_discovered_devices = {
                    d for d in ie.available_devices
                    if not d.lower().startswith("cpu")
                }
            except ModuleNotFoundError:
                logging.warning("OpenVINO library not found. "
                                "OpenVINO devices will not be discovered. ")
                openvino_discovered_devices = set()
            _devices = list(tf_discovered_devices
                            | openvino_discovered_devices)
    return _devices


class DeviceMapper:
    def __init__(self, filter_func: Callable[[List[str]], List[str]]):
        """The filter will take in a list of devices formatted as
        ["CPU:0", "CPU:1", "GPU:0", "GPU:1"], etc and output a filtered list of
        devices.
        """
        self.filter_func = filter_func

    @staticmethod
    def map_to_all_gpus(cpu_fallback=True) -> 'DeviceMapper':
        def filter_func(devices):
            gpu_devices = [d for d in devices if d.startswith("GPU:")]
            if not gpu_devices and cpu_fallback:
                return ["CPU:0"]
            return gpu_devices

        return DeviceMapper(filter_func=filter_func)

    @staticmethod
    def map_to_single_cpu() -> 'DeviceMapper':
        def filter_func(devices):
            return [next(d for d in devices if d.startswith("CPU:"))]

        return DeviceMapper(filter_func=filter_func)

    @staticmethod
    def map_to_openvino_devices():
        """Intelligently load capsules onto available OpenVINO-compatible
        devices.

        Since support for OpenVINO devices is experimental, there is a
        temporary environment variable being added to whitelist devices
        specifically. This variable will be deprecated and removed after a
        short testing period.

        The device "CPU" is _always_ allowed and always loaded onto and cannot
        be excluded.

        Here are the cases:
        ['CPU:0', 'HDDL', ...] =>  ["MULTI:CPU,HDDL"]

        ['CPU:0'] => ["CPU"]
            Always load onto CPU.
        """

        def filter_func(devices):

            devices_by_priority = os.environ.get(
                "OPENVINO_DEVICE_PRIORITY",
                "CPU,HDDL").split(",")
            load_to_devices = []

            for device in devices_by_priority:
                for existing_device in devices:
                    if existing_device.lower().startswith(device.lower()):
                        load_to_devices.append(device)

            if len(load_to_devices) > 1:
                return ["MULTI:" + ",".join(load_to_devices)]
            else:
                return load_to_devices

        return DeviceMapper(filter_func=filter_func)



import re
import threading
from typing import List, Optional


def verify_all_threads_closed(allowable_threads: Optional[List[str]] = None):
    """A convenient function to throw an error if all threads are not closed.

    :param allowable_threads: A list of regular expressions that match to
        thread names that are allowed to stay open
    """
    if allowable_threads is None:
        allowable_threads = []
    allowable_threads += [
        r"pydevd\.Writer",
        r"pydevd\.Reader",
        r"pydevd\.CommandThread",
        r"profiler\.Reader",
        r"MainThread",
    ]

    open_threads = []

    for thread in threading.enumerate():
        matched = False
        for name_pattern in allowable_threads:
            if re.match(name_pattern, thread.name):
                matched = True
                break

        if not matched:
            open_threads.append(thread)

    if len(open_threads) != 0:
        raise EnvironmentError(
            "Not all threads were shut down! Currently running threads: " +
            str(open_threads))



from typing import BinaryIO

from Cryptodome.Cipher import AES


def encrypt(password: str, data: bytes, dest: BinaryIO) -> None:
    """Encrypt the given source to dest.

    :param password: The password to encrypt with
    :param data: Data to encrypt
    :param dest: A file-like object to write results to
    """
    cipher = AES.new(password.encode('utf-8'), AES.MODE_EAX)
    ciphertext, tag = cipher.encrypt_and_digest(data)

    [dest.write(x) for x in (cipher.nonce, tag, ciphertext)]


def decrypt(password: str, data: bytes) -> bytes:
    """Decrypt the given bytes.

    :param password: The password to decrypt with
    :param data: Data to decrypt
    :return: The resulting bytes
    """
    nonce = data[:16]
    tag = data[16:32]
    ciphertext = data[32:]

    cipher = AES.new(password.encode("utf-8"), AES.MODE_EAX, nonce)
    return cipher.decrypt_and_verify(ciphertext, tag)



import logging
import random
from abc import ABC, abstractmethod
from threading import RLock
from typing import Dict, List, Optional

import numpy as np

from .backend import BaseBackend
from .caching import cache
from .device_mapping import DeviceMapper, get_all_devices
from .node_description import DETECTION_NODE_TYPE, NodeDescription
from .options import OPTION_TYPE, Option
from .stream_state import BaseStreamState


class BaseCapsule(ABC):
    """An abstract base class that all capsules must subclass. Defines the
    interface that capsules are expected to implement.

    A class that subclasses from this class is expected to be defined in a
    capsule.py file in a capsule.
    """

    backends: Optional[List[BaseBackend]] = None
    """A list of the backends the capsule has initialized"""

    backends_lock = RLock()
    """Lock the list of backends whenever it's being accessed"""

    def __init__(self, capsule_files: Dict[str, bytes], inference_mode=True):
        """
        :param capsule_files: A dict of {"file_name": FILE_BYTES} of the files
            that were found and loaded in the capsule
        :param inference_mode: If True, the model will be loaded and the
            backends will start for it. If False, the capsule will never be
            able to run inference, but it will still have it's various readable
            attributes.
        """

        if inference_mode:
            # We use type(self) to avoid having capsules need to put
            # staticmethod on each of their backend loaders.
            # self.backend_loader is a function not a method because we usually
            # use the following system for creating subclasses:
            #
            # >>> class CoolCapsule(BaseCapsule):
            # ...     ...
            # ...     backend_loader = lambda: some_func(...)
            #
            # This prevents self from being passed as the first (and wrong)
            # argument. We may want to change this in the future.
            load_backend = lambda device: type(self).backend_loader(
                capsule_files=capsule_files,
                device=device)

            # Initialize the backends on the capsules requested devices
            # for example: GPU:0, GPU:1, CPU:0, MYRIAD, HDDL
            all_devices = get_all_devices()
            devices_to_load_to = self.device_mapper.filter_func(all_devices)

            # Throw an error if the capsule couldn't load
            if not devices_to_load_to:
                message = f"The capsule was not able to find any valid " \
                          f"devices to initialize on! The devices " \
                          f"discovered on this machine are: {all_devices}."
                raise EnvironmentError(message)

            logging.info(f"Loading capsule {self.name} "
                         f"onto devices {', '.join(devices_to_load_to)}")
            self.backends = list(map(load_backend, devices_to_load_to))

        # Keep a dictionary for keeping track of different StreamState objs
        self._stream_states: Dict[int, BaseStreamState] = {}
        """{stream_id: BaseStreamState}"""

        self._stream_state_lock = RLock()

    def __repr__(self):
        rep = f"Capsule(name={self.name}, " \
              f"input={self.input_type}, " \
              f"output={self.output_type})"
        return rep

    def __del__(self):
        self.close()

    def process_frame(self,
                      frame: np.ndarray,
                      detection_node: DETECTION_NODE_TYPE,
                      options: Dict[str, OPTION_TYPE],
                      state: BaseStreamState) \
            -> DETECTION_NODE_TYPE:
        """
        Find the Backend that has a BatchExecutor with the least amount of
        images in the pipeline, and run with that Backend. In multi-GPU
        scenarios, this is the logic that allows even usage across all GPUs.
        """
        with self.backends_lock:
            # Pick the backend with the lowest current 'workload' and
            # send the image there for inferencing.
            # Note the inplace shuffle, but this should be okay, esp. w/ lock
            random.shuffle(self.backends)
            laziest_backend: BaseBackend \
                = min(self.backends, key=lambda backend: backend.workload)

        return laziest_backend.process_frame(
            frame, detection_node, options, state)

    def get_state(self, stream_id: int) -> BaseStreamState:
        """Get or create the StreamState for this stream_id and return it."""
        with self._stream_state_lock:
            # Adds stream state if stream_id is not already a key
            return self._stream_states.setdefault(stream_id,
                                                  self.stream_state())

    def clean_up(self, stream_id: int) -> None:
        """Dereference the StreamState that relates to this stream_id"""
        with self._stream_state_lock:
            if stream_id in self._stream_states:
                del self._stream_states[stream_id]

    def close(self) -> None:
        """De-initializes all backends."""
        if self.backends is not None:
            with self.backends_lock:
                for backend in self.backends:
                    backend.close()
                self.backends = None

    @property
    @cache
    def capability(self) -> NodeDescription:
        """Returns a NodeDescription showing the difference between the input
        and output NodeDescriptions for this capsule. This tells the caller
        what service this capsule provides. For instance, a capsule that
        classifies a person's gait would return the following:

        >>> NodeDescription(
        ...     size=NodeDescription.Size.ALL,
        ...     attributes={"Gait": ["walking", "running"]})

        The fact that the capsule's input is a person is not shown in this
        value because person detection is not something this capsule provides.
        """
        return self.input_type.difference(self.output_type)

    @property
    @cache
    def default_options(self):
        """Return a dict of key->value where the value is the default value
        for this capsules options"""
        return {key: val.default for key, val in self.options.items()}

    # These things are to be filled in by a subclassing capsule
    @property
    def name(self) -> str:
        """The name of the capsule. This value uniquely defines the capsule and
        cannot be shared by other capsules.

        By convention, the capsule's name is prefixed by some short description
        of the role it plays ("detector", "recognizer", etc) followed by the
        kind of data it relates to ("person", "face", etc) and, if necessary,
        some differentiating factor ("fast", "close_up", "retail", etc). For
        example, a face detector that is optimized for quick inference would
        be named "detector_face_fast".
        """
        raise NotImplementedError

    @property
    def version(self) -> int:
        """The version of the capsule.

        When should you bump the version of a capsule? When:
            - You've changed the usage of existing capsule options
            - You've changed the model or algorithm
            - You've changed the input/output node descriptions

        When shouldn't you bump the version of a capsule? When:
            - You only did code restructuring
            - You've updated the capsule to work with a newer API version
            - You've added (but not removed or changed previous) capsule
              options

        In summary, the version is most useful for differentiating a capsule
        from its previous versions.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """A human-readable description of what the capsule does."""
        return ""

    @property
    def device_mapper(self) -> DeviceMapper:
        """A device mapper contains a single field, ``filter_func``, which is a
        function that takes in a list of all available device strings and
        returns a list of device strings that are compatible with this capsule.
        """
        return DeviceMapper.map_to_all_gpus()

    @staticmethod
    @abstractmethod
    def backend_loader(capsule_files: Dict[str, bytes], device: str) \
            -> "BaseBackend":
        """A function that creates a backend for this capsule.

        :param capsule_files: Provides access to all files in the capsule. The
            keys are file names and the values are ``bytes``.
        :param device: A string specifying the device that this backend should
            use. For example, the first GPU device is specified as "GPU:0".
        """
        raise NotImplementedError

    @property
    def input_type(self) -> NodeDescription:
        """Describes the types of DetectionNodes that this capsule takes in as
        input.
        """
        raise NotImplementedError

    @property
    def output_type(self) -> NodeDescription:
        """Describes the types of DetectionNodes that this capsule produces as
        output.
        """
        raise NotImplementedError

    @property
    def options(self) -> Dict[str, Option]:
        """A list of zero or more options that can be configured at runtime."""
        return {}

    @property
    def stream_state(self):
        """(Optional) An instance of this object will be created for every new
        video stream that a capsule is run on, and de-initialized when that
        stream is deleted. It is intended to be overridden by capsules that
        have stateful operations across a single stream.
        """
        return BaseStreamState



from abc import ABC, abstractmethod
from typing import List, Optional, Union


OPTION_TYPE = Union[int, float, bool, str]


def _check_number_range(value, min_val, max_val):
    error = ValueError(f"Value {value} is not in "
                       f"{min_val} <= x <= {max_val}")

    if min_val is not None and value < min_val:
        raise error
    if max_val is not None and value > max_val:
        raise error


class Option(ABC):
    """A class that all options subclass. Added mostly for documentation an
    type hinting purposes.
    """

    def __init__(self, default, description):
        self.default = default
        self.description: Optional[str] = description
        self.check(default)

    @abstractmethod
    def check(self, value):
        """Checks the given value to see if it fits this option.

        :param value: The value to check
        :raises ValueError: If the value does not fit the option
        """
        pass


class FloatOption(Option):
    """A capsule option that holds a floating point value with defined
    boundaries.
    """

    def __init__(self, *, default: float,
                 min_val: Optional[float],
                 max_val: Optional[float],
                 description: Optional[str] = None):
        """
        :param default: The default value of this option
        :param min_val: The minimum allowed value for this option, inclusive,
            or None for no lower limit
        :param max_val: The maximum allowed value for this option, inclusive,
            or None for no upper limit
        :param description: The description for this option
        """
        self.min_val = min_val
        self.max_val = max_val

        super().__init__(default=default, description=description)

    def check(self, value):
        if not isinstance(value, float):
            raise ValueError(f"Expecting type float, got {type(value)}")
        _check_number_range(value, self.min_val, self.max_val)


class IntOption(Option):
    """A capsule option that holds an integer value.
    """

    def __init__(self, *, default: int,
                 min_val: Optional[int],
                 max_val: Optional[int],
                 description: Optional[str] = None):
        """
        :param default: The default value of this option
        :param min_val: The minimum allowed value for this option, inclusive,
            or None for no lower limit
        :param max_val: The maximum allowed value for this option, inclusive,
            or None for no upper limit
        :param description: The description for this option
        """
        self.min_val = min_val
        self.max_val = max_val

        super().__init__(default=default, description=description)

    def check(self, value):
        if not isinstance(value, int):
            raise ValueError(f"Expecting type int, got {type(value)}")
        _check_number_range(value, self.min_val, self.max_val)


class EnumOption(Option):
    """A capsule option that holds a choice from a discrete set of string
    values.
    """

    def __init__(self, *, default: str, choices: List[str],
                 description: Optional[str] = None):
        """
        :param default: The default value of this option
        :param choices: A list of all valid values for this option
        :param description: The description for this option
        """
        assert len(choices) > 0

        self.choices = choices

        super().__init__(default=default, description=description)

    def check(self, value):
        if not isinstance(value, str):
            raise ValueError(f"Expecting type float, got {type(value)}")

        if value not in self.choices:
            raise ValueError(f"Value {value} is not one of {self.choices}")


class BoolOption(Option):
    """A capsule option that holds an boolean value."""

    def __init__(self, *, default: bool,
                 description: Optional[str] = None):
        """
        :param default: The default value of this option
        :param description: The description for this option
        """
        super().__init__(default=default, description=description)

    def check(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"Expecting type bool, got {type(value)}")


def check_option_values(capsule, option_vals):
    """Checks the given option values against the capsule's available options to
    make sure they are of the right name, type, and fit within the constraints.
    """
    for name, val in option_vals.items():
        if name not in capsule.options:
            raise ValueError(f"'{name}' is not a valid option for this capsule")
        option = capsule.options[name]
        try:
            option.check(val)
        except ValueError as e:
            raise ValueError(f"For option {name}: {e}")


"""Options that detector capsules tend to have
to avoid breaking API too often, make these options be short, with minimal 
keys. Try to add a name that is descriptive to the use case of the capsule.

If an existing one is changed, it is the responsibility of the person
changing to verify that all capsules have been updated. This is considered
a breaking capsule API change if the key is changed.
"""

common_detector_options = {
    "threshold": FloatOption(
        default=0.5,
        min_val=0.0,
        max_val=1.0,
        description="The confidence threshold for the detector to return a "
                    "detection for an object. Lower => more detections, "
                    "higher means fewer but more accurate detections."),
}



import configparser
import hashlib
import re
import sys
from io import BytesIO
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, Optional, Union
from zipfile import ZipFile

from vcap import BaseCapsule, BaseBackend, BaseStreamState, NodeDescription
from vcap.loading.errors import IncompatibleCapsuleError, InvalidCapsuleError

from .crypto_utils import decrypt
from .import_hacks import ZipFinder
from .packaging import CAPSULE_FILE_NAME, META_FILE_NAME

MAJOR_COMPATIBLE_VERSION = 0
MINOR_COMPATIBLE_VERSION = 3
"""The capsule version that this version of the library supports."""

MAJOR_MINOR_SEMVER_PATTERN = re.compile(r"([0-9]+)\.([0-9]+)")


def load_capsule_from_bytes(data: bytes,
                            source_path: Optional[Path] = None,
                            key: Optional[str] = None,
                            inference_mode: bool = True) -> BaseCapsule:
    """Loads a capsule from the given bytes.

    :param data: The data of the capsule
    :param source_path: The path to the capsule's source code, if it's
        available at runtime
    :param key: The AES key to decrypt the capsule with, or None if the capsule
        is not encrypted
    :param inference_mode: If True, the backends for this capsule will be
        started. If False, the capsule will never be able to run inference, but
        it will still have it's various readable attributes.
    :return: The loaded capsule object
    """
    module_name = capsule_module_name(data)
    if source_path is None:
        # The source is unavailable. Use a dummy path
        source_path = Path("/", module_name)

    if key is not None:
        # Decrypt the capsule into its original form, a zip file
        data = decrypt(key, data)

    file_like = BytesIO(data)
    loaded_files = {}

    code = None
    with ZipFile(file_like, "r") as capsule_file:
        if CAPSULE_FILE_NAME not in capsule_file.namelist():
            raise InvalidCapsuleError(f"Missing file {CAPSULE_FILE_NAME}")

        if META_FILE_NAME not in capsule_file.namelist():
            raise InvalidCapsuleError(f"Missing file {META_FILE_NAME}")

        for name in capsule_file.namelist():
            if name == CAPSULE_FILE_NAME:
                # Every capsule has a capsule.py file defining the capsule's
                # behavior
                code = capsule_file.read(CAPSULE_FILE_NAME)
            else:
                # Load all other files as well
                loaded_files[name] = capsule_file.read(name)

        # Read the meta.conf and get the OpenVisionCapsules API compatibility
        # version
        meta_conf = configparser.ConfigParser()
        meta_conf.read_string(loaded_files[META_FILE_NAME].decode("utf-8"))
        compatibility_version = meta_conf["about"]["api_compatibility_version"]

        match = MAJOR_MINOR_SEMVER_PATTERN.fullmatch(compatibility_version)
        if match is None:
            raise InvalidCapsuleError(
                f"Invalid API compatibility version format "
                f"'{compatibility_version}'. Version must be in the format "
                f"'[major].[minor]'.")
        try:
            major, minor = map(int, (match[1], match[2]))
        except ValueError:
            raise InvalidCapsuleError(
                f"Compatibility versions must be numbers, got "
                f"{major}.{minor}.")
        if major != MAJOR_COMPATIBLE_VERSION:
            raise IncompatibleCapsuleError(
                f"The capsule is not compatible with this software. The "
                f"capsule's OpenVisionCapsules required major version is "
                f"{major} but this software uses OpenVisionCapsules "
                f"{MAJOR_COMPATIBLE_VERSION}.{MINOR_COMPATIBLE_VERSION}.")
        if minor > MINOR_COMPATIBLE_VERSION:
            raise IncompatibleCapsuleError(
                f"The capsule requires a version of OpenVisionCapsules "
                f"that is too new for this software. The capsule requires at "
                f"least version {major}.{minor} but this software uses "
                f"OpenVisionCapsules "
                f"{MAJOR_COMPATIBLE_VERSION}.{MINOR_COMPATIBLE_VERSION}.")

        # With the capsule's code loaded, initialize the object
        capsule_module = ModuleType(module_name)

        # Allow the capsule.py to import other files in the capsule
        sys.meta_path.insert(
            1, ZipFinder(capsule_file, source_path, module_name))

        try:
            # Run the capsule
            compiled = compile(code, source_path / CAPSULE_FILE_NAME, "exec")
            exec(compiled, capsule_module.__dict__)
        except Exception as e:
            raise InvalidCapsuleError(
                f"Could not execute the code in the capsule!\n"
                f"Error: {e}")
        finally:
            # Remove custom import code
            sys.meta_path.pop(1)

    # noinspection PyUnresolvedReferences
    new_capsule: BaseCapsule = capsule_module.Capsule(
        capsule_files=loaded_files,
        inference_mode=inference_mode)

    try:
        _validate_capsule(new_capsule)
    except InvalidCapsuleError as e:
        new_capsule.close()
        raise e

    return new_capsule


def load_capsule(path: Union[str, Path],
                 source_path: Optional[Path] = None,
                 key: Optional[str] = None,
                 inference_mode: bool = True) -> BaseCapsule:
    """Load a capsule from the filesystem.

    :param path: The path to the capsule file
    :param source_path: The path to the capsule's source code, if it's
        available at runtime
    :param key: The AES key to decrypt the capsule with, or None if the capsule
        is not encrypted
    :param inference_mode: If True, the backends for this capsule will be
        started. If False, the capsule will never be able to run inference, but
        it will still have it's various readable attributes.
    """
    path = Path(path)

    if source_path is None:
        # Set the default source path to a directory alongside the capsule file
        source_path = path.absolute().with_suffix("")

    return load_capsule_from_bytes(
        data=path.read_bytes(),
        source_path=source_path,
        key=key,
        inference_mode=inference_mode,
    )


def _validate_capsule(capsule: BaseCapsule):
    """This will try calling different attributes on a capsule to make sure
    they are there. This is based on the API compatibility version.

    :raises: InvalidCapsuleError
    """

    def check_arg_names(func: Callable,
                        correct: List[str],
                        ignore: Optional[List[str]] = None) \
            -> bool:
        """Return False if a function has the wrong argument names. Return
        true if they are correct.
        Usage:
        >>> def my_func(self, frame, detection_node):
        ...     pass
        >>> check_arg_names(my_func, ['self'], ['frame', 'detection_node'])
        True
        """
        # noinspection PyUnresolvedReferences
        code = func.__code__

        ignore = [] if ignore is None else ignore
        all_var_names = code.co_varnames
        arg_names = all_var_names[:code.co_argcount]
        filtered = [n for n in arg_names if n not in ignore]
        return filtered == correct

    try:
        # If the ASCII flag is used, only [a-zA-Z0-9_] is matched for \w.
        # Otherwise, some other Unicode characters can be matched, depending on
        # the locale. We don't want that.
        if re.fullmatch(r"\w+", capsule.name, flags=re.ASCII) is None:
            raise InvalidCapsuleError(
                "Capsule names must only contain alphanumeric characters and "
                "underscores",
                capsule.name)

        # Validate the capsule class attributes
        _validate_capsule_field(capsule, "name", capsule.name, str)
        _validate_capsule_field(capsule,
                                "backend_loader",
                                capsule.backend_loader,
                                callable)
        _validate_capsule_field(capsule, "version", capsule.version, int)
        _validate_capsule_field(capsule,
                                "input_type",
                                capsule.input_type,
                                NodeDescription)
        _validate_capsule_field(capsule,
                                "output_type",
                                capsule.input_type,
                                NodeDescription)
        _validate_capsule_field(capsule, "options", capsule.options, dict)

        # Make sure that certain things are NOT attributes (we don't want
        # accidental leftover code from previous capsule versions)
        unwanted_attributes = ["backend_config"]
        for unwanted_attr in unwanted_attributes:
            try:
                # This should throw an attribute error
                capsule.__getattribute__(unwanted_attr)
                raise InvalidCapsuleError(
                    f"The capsule has leftover attributes from a previous "
                    f"OpenVisionCapsules API version. Attribute name: "
                    f"{unwanted_attr}",
                    capsule.name)
            except AttributeError:
                pass

        # Validate the capsule's backend_loader takes the right args
        correct_args = ["capsule_files", "device"]
        is_correct = check_arg_names(func=capsule.backend_loader,
                                     correct=correct_args)
        if not is_correct:
            raise InvalidCapsuleError(
                f"The capsule's backend_loader has invalid arguments!\n"
                f"The arguments must be {correct_args}",
                capsule.name)

        # Validate the backend class attributes
        if capsule.backends is not None:
            backend = capsule.backends[0]
            _validate_backend_field(capsule,
                                    "batch_predict",
                                    backend.batch_predict,
                                    callable)
            _validate_backend_field(capsule,
                                    "process_frame",
                                    backend.process_frame,
                                    callable)
            _validate_backend_field(capsule,
                                    "close",
                                    backend.close,
                                    callable)
            if not isinstance(capsule.backends[0], BaseBackend):
                raise InvalidCapsuleError(
                    f"The capsule's backend field must be a class that "
                    f"subclasses {BaseBackend.__name__}",
                    capsule.name)

        # Validate the stream state
        stream_state = capsule.stream_state
        if not issubclass(stream_state, BaseStreamState):
            raise InvalidCapsuleError(
                f"The capsule's stream_state field must be a subclass of "
                f"{BaseStreamState.__name__}, got {stream_state}")

        # Validate that if the capsule is an encoder, it has a threshold option
        if capsule.capability.encoded:
            if "recognition_threshold" not in capsule.options.keys():
                raise InvalidCapsuleError(
                    "This capsule can encode, but doesn't have a option "
                    "called \"recognition_threshold\"!",
                    capsule.name)
    except InvalidCapsuleError:
        # Don't catch this exception type in the "except Exception" below
        raise
    except AttributeError as e:
        message = f"The capsule did not describe the necessary attributes. " \
                  f"Error: {e}"
        raise InvalidCapsuleError(message, capsule.name)
    except Exception as e:
        message = f"This capsule is invalid for unknown reasons. Error: {e}"
        raise InvalidCapsuleError(message, capsule.name)


def capsule_module_name(data: bytes) -> str:
    """Creates a unique module name for the given capsule bytes"""
    hash_ = hashlib.sha256(data).hexdigest()
    return f"capsule_{hash_}"


_TYPE_CALLABLE = Union[type, callable]


def _validate_capsule_field(capsule: BaseCapsule,
                            name: str,
                            value: Any,
                            type_: _TYPE_CALLABLE) -> None:
    if type_ is callable:
        if not callable(value):
            raise InvalidCapsuleError(
                f"The capsule has an invalid internal configuration!\n"
                f"Capsule field {name} must be callable, got {type(value)}",
                capsule.name)
    elif not isinstance(value, type_):
        raise InvalidCapsuleError(
            f"The capsule has an invalid internal configuration!\n"
            f"Capsule field {name} must be of type {type_}, got {type(value)}",
            capsule.name)


def _validate_backend_field(capsule: BaseCapsule,
                            name: str,
                            value: Any,
                            type_: _TYPE_CALLABLE) -> None:
    if type_ is callable:
        if not callable(value):
            raise InvalidCapsuleError(
                f"The capsule's backend has an invalid configuration!\n"
                f"Backend field {name} must be callable, got {type(value)}",
                capsule.name)
    elif not isinstance(value, type_):
        raise InvalidCapsuleError(
            f"The capsule's backend has an invalid configuration!\n"
            f"Backend field {name} must be of type {type_}, got {type(value)}",
            capsule.name)



class CapsuleLoadError(Exception):
    """Raised when there was an error loading a capsule."""

    def __init__(self, message, capsule_name=None):
        self.capsule_name = capsule_name
        if capsule_name is not None:
            message = f"Capsule {capsule_name}: {message}"

        super().__init__(message)


class IncompatibleCapsuleError(CapsuleLoadError):
    """This is thrown when a capsule is of an incompatible API version for this
    version of the library.
    """


class InvalidCapsuleError(CapsuleLoadError):
    """This is thrown when a capsule has the correct API compatibility version,
    but it still doesn't have the correct information on it.
    """



import warnings
from typing import Optional
from pkg_resources import parse_version

import functools

from . import __version__


def deprecated(message: str = "",
               remove_in: Optional[str] = None,
               current_version=__version__):
    """Mark a function as deprecated
    :param message: Extra details to be added to the warning. For example,
    the details may point users to a replacement method.
    :param remove_in: The version when the decorated method will be removed.
    The default is None, specifying that the function is not currently planned
    to be removed.
    :param current_version: Defaults to the vcaps current version.
    """

    def wrapper(function):
        warning_msg = f"'{function.__qualname__}' is deprecated "
        warning_msg += (f"and is scheduled to be removed in {remove_in}. "
                        if remove_in is not None else ". ")
        warning_msg += str(message)

        # Only show the DeprecationWarning on the first call
        warnings.simplefilter("once", DeprecationWarning, append=True)

        @functools.wraps(function)
        def inner(*args, **kwargs):
            if remove_in is not None and \
                    parse_version(current_version) >= parse_version(remove_in):
                raise DeprecationWarning(warning_msg)
            else:
                warnings.warn(warning_msg, category=DeprecationWarning)
            return function(*args, **kwargs)

        return inner

    return wrapper



import queue
from concurrent.futures import Future
from queue import Queue
from threading import Thread
from typing import Any, Callable, Iterable, List, NamedTuple, Optional

from vcap import deprecated


class _Request(NamedTuple):
    """Used by BatchExecutor to keep track of requests and their respective
    future objects.
    """

    future: Future
    """The Future object for the BatchExecutor to return the output"""

    input_data: Any
    """A unit of input data expected by the batch_fn."""


class BatchExecutor:
    """Feeds jobs into batch_fn in batches, returns results through Futures.

    This class simplifies centralizing work from a multitude of sources and
    running that work in a batched predict function, then returning that
    work to the respective Futures.
    """

    def __init__(self,
                 batch_fn: Callable[[List[Any]], Iterable[Any]],
                 max_batch_size=40,
                 num_workers: int = 1):
        """Initialize a new BatchExecutor

        :param batch_fn: A function that takes in a list of inputs and iterates
        the outputs in the same order as the inputs.
        :param max_batch_size: The maximum length of list to feed to batch_fn
        :param num_workers: How many workers should be calling batch_fn
        """
        self.batch_fn = batch_fn
        self.max_batch_size = max_batch_size
        self._request_queue: Queue[_Request] = Queue()
        self.workers = [Thread(target=self._worker,
                               daemon=True,
                               name="BatchExecutorThread")
                        for _ in range(num_workers)]

        # The number of images currently in the work queue or being processed
        self._num_imgs_being_processed: int = 0

        self._running: bool = True

        for worker in self.workers:
            worker.start()

    @property
    def total_imgs_in_pipeline(self) -> int:
        return self._request_queue.qsize() + self._num_imgs_being_processed

    def submit(self, input_data: Any, future: Future = None) -> Future:
        """Submits a job and returns a Future that will be fulfilled later."""
        future = future or Future()

        self._request_queue.put(_Request(
            future=future,
            input_data=input_data))
        return future

    def _on_requests_ready(self, batch: List[_Request]) -> None:
        """Push inputs through the given prediction backend

        :param batch: A list of requests to work on
        """
        # Extract the futures from the requests
        inputs: List[Any] = [req.input_data for req in batch]
        futures: List[Future] = [req.future for req in batch]

        # Route the results to each request
        try:
            for prediction in self.batch_fn(inputs):
                # Popping the futures ensures that if an error occurs, only
                # the futures that haven't had a result set will have
                # set_exception called
                futures.pop(0).set_result(prediction)
        except BaseException as exc:
            # Catch exceptions and pass them to the futures, similar to the
            # ThreadPoolExecutor implementation:
            # https://github.com/python/cpython/blob/91e93794/Lib/concurrent/futures/thread.py#L51
            for future in futures:
                future.set_exception(exc)

    def _worker(self):
        self._running = True

        while self._running:
            # Get a new batch
            batch = self._get_next_batch()

            # If no batch was able to be retrieved, restart the loop
            if batch is None:
                continue

            # Check to make sure the thread isn't trying to end
            if not self._running:
                break

            self._num_imgs_being_processed += len(batch)
            self._on_requests_ready(batch)
            self._num_imgs_being_processed -= len(batch)

        self._running = False

    def _get_next_batch(self) -> Optional[List[_Request]]:
        """A helper function to help make the main thread loop more readable
        :returns: A non-empty list of collected items, or None if the worker is
                  no longer running (i.e. self._continue == False)
        """
        batch: List[_Request] = []
        while len(batch) < self.max_batch_size:
            # Check if there's a new request
            try:
                # Try to get a new request. Have a timeout to check if closing
                new_request = self._request_queue.get(timeout=.1)
            except queue.Empty:
                # If the thread is being requested to close, exit early
                if not self._running:
                    return None

                # Wait for requests again
                continue

            batch.append(new_request)

            # If the request queue is now empty, let worker run everything in
            # the batch
            if self._request_queue.empty():
                break

        return batch

    def close(self) -> None:
        """Stop the BatchExecutor gracefully."""
        self._running = False
        for worker in self.workers:
            worker.join()



import abc
from concurrent.futures import Future
from typing import Any, Dict, List

import numpy as np

from vcap.batch_executor import BatchExecutor
from vcap.node_description import DETECTION_NODE_TYPE
from vcap.options import OPTION_TYPE
from vcap.stream_state import BaseStreamState


class BaseBackend(abc.ABC):
    """An object that provides low-level prediction functionality for batches
    of frames.
    """

    def __init__(self):
        self._batch_executor = BatchExecutor(self.batch_predict)

    def send_to_batch(self, input_data: Any) -> Future:
        """Sends the given object to the batch_predict method for processing.
        This call does not block. Instead, the result will be provided on the
        returned Future. The batch_predict method must be overridden on the
        backend this method is being called on.

        :param input_data: The input object to send to batch_predict
        :return: A Future where results will be stored
        """
        return self._batch_executor.submit(input_data)

    @property
    def workload(self) -> float:
        """Returns a unit representing the amount of 'work' being processed
        This value is comparable only by backends of the same capsule, and
        is intended to give the scheduler the ability to pick the least busy
        backend.
        """
        return self._batch_executor.total_imgs_in_pipeline

    @abc.abstractmethod
    def process_frame(self,
                      frame: np.ndarray,
                      detection_node: DETECTION_NODE_TYPE,
                      options: Dict[str, OPTION_TYPE],
                      state: BaseStreamState) \
            -> DETECTION_NODE_TYPE:
        """A method that does the pre-processing, inference, and postprocessing
        work for a frame.

        If the capsule uses an algorithm that benefits from batching,
        this method may call ``self.send_to_batch``, which will asynchronously
        send work out for batching. Doing so requires that the
        ``batch_predict`` method is overridden.

        :param frame: A numpy array representing a frame. It is of shape
            (height, width, num_channels) and the frames come in BGR order.
        :param detection_node: The detection_node type as specified by the
            ``input_type``
        :param options: A dictionary of key (string) value pairs. The key is
            the name of a capsule option, and the value is its configured value
            at the time of processing. Capsule options are specified using the
            ``options`` field in the Capsule class.
        :param state: This will be a StreamState object of the type specified
            by the ``stream_state`` attribute on the Capsule class. If no
            StreamState object was specified, a simple BaseStreamState object
            will be passed in. The StreamState will be the same object for all
            frames in the same video stream.
        """

    def batch_predict(self, input_data_list: List[Any]) -> List[Any]:
        """This method takes in a batch as input and provides a list of result
        objects of any type as output. What the result objects are will depend
        on the algorithm  being defined, but the number of prediction objects
        returned _must_ match the number of video frames provided as input.

        :param input_data_list: A list of objects. Whatever the model requires
            for each frame.
        """
        raise NotImplementedError(
            "Attempt to do batch prediction on a Backend that does not have "
            "the batch_predict method defined. Did you call send_to_batch on "
            "a backend that does not override batch_predict?")

    def close(self) -> None:
        """De-initializes the backend. This is called when the capsule is being
        unloaded. This method should be overridden by any Backend that needs
        to release resources or close other threads.

        The backend will stop receiving frames before this method is
        called, and will not receive frames again.
        """
        self._batch_executor.close()



#!/usr/bin/env python3
import os
from pathlib import Path

from setuptools import setup, find_namespace_packages

# Get package version/metadata
about = {}
exec(Path("vcap/version.py").read_text(), about)

test_packages = ["pytest", "mock"]

PRE_RELEASE_SUFFIX = os.environ.get("PRE_RELEASE_SUFFIX", "")

setup(
    name='vcap',
    description="A library for creating OpenVisionCapsules in Python",
    author="Aotu.ai",
    packages=find_namespace_packages(include=["vcap*"]),
    version=about["__version__"] + PRE_RELEASE_SUFFIX,

    install_requires=[
        "pycryptodomex==3.9.7",
        "scipy==1.4.1",
        "scikit-learn==0.22.2",
        "numpy>=1.16,<2",
        "tensorflow-gpu==1.15.4",
    ],
    extras_require={
        "tests": test_packages,
        "easy": ["opencv-python-headless~=4.1.2"],
    },
    tests_require=test_packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
